using System;

namespace ThirdLeftChild {
  public static class MaybeHelpers {
    public static Maybe<T> ToMaybe<T>(this T val) {
      return new Maybe<T>(val);
    }

    public static Maybe<T> ToNotNullMaybe<T>(this T val) where T : class {
      return val != null ? val.ToMaybe( ) : Maybe<T>.Empty;
    }

    public static Maybe<R> SelectMany<T, R>(this Maybe<T> m, Func<T, Maybe<R>> g) {
      return m.Bind(g);
    }

    public static Maybe<R> SelectMany<T, TCollection, R>(
          this Maybe<T> source,
          Func<T, Maybe<TCollection>> collectionSelector,
          Func<T, TCollection, R> resultSelector) {
      if (source.IsEmpty)
        return Maybe<R>.Empty;
      var x = source.Bind(collectionSelector);
      if (x.IsEmpty)
        return Maybe<R>.Empty;
      var r = resultSelector(source.Value, x.Value);
      return r.ToMaybe( );
    }
  }
}
